package com.bakdata.conquery.apiv1.query.concept.specific.external;


import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import com.bakdata.conquery.apiv1.query.CQElement;
import com.bakdata.conquery.io.cps.CPSType;
import com.bakdata.conquery.io.jackson.View;
import com.bakdata.conquery.models.common.CDateSet;
import com.bakdata.conquery.models.config.IdColumnConfig;
import com.bakdata.conquery.models.error.ConqueryError;
import com.bakdata.conquery.models.identifiable.ids.specific.ManagedExecutionId;
import com.bakdata.conquery.models.identifiable.mapping.EntityIdMap;
import com.bakdata.conquery.models.query.QueryExecutionContext;
import com.bakdata.conquery.models.query.QueryPlanContext;
import com.bakdata.conquery.models.query.QueryResolveContext;
import com.bakdata.conquery.models.query.RequiredEntities;
import com.bakdata.conquery.models.query.queryplan.ConceptQueryPlan;
import com.bakdata.conquery.models.query.queryplan.QPNode;
import com.bakdata.conquery.models.query.queryplan.aggregators.specific.ConstantValueAggregator;
import com.bakdata.conquery.models.query.queryplan.specific.ExternalNode;
import com.bakdata.conquery.models.query.resultinfo.ResultInfo;
import com.bakdata.conquery.models.query.resultinfo.SimpleResultInfo;
import com.bakdata.conquery.models.types.ResultType;
import com.bakdata.conquery.util.DateReader;
import com.bakdata.conquery.util.io.IdColumnUtil;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonView;
import com.google.common.collect.Streams;
import io.dropwizard.validation.ValidationMethod;
import jakarta.validation.constraints.NotEmpty;
import jakarta.validation.constraints.NotNull;
import lombok.AccessLevel;
import lombok.Data;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.extern.slf4j.Slf4j;

/**
 * Allows uploading lists of entities.
 */
@Slf4j
@CPSType(id = "EXTERNAL", base = CQElement.class)
@NoArgsConstructor
public class CQExternal extends CQElement {

	private static final String FORMAT_EXTRA = "EXTRA";

	/**
	 * Describes the format of {@code values}, how to extract data from each row:
	 * <p>
	 * - Must contain at least one of {@link IdColumnConfig#getIds()}.
	 * - May contain names of {@link DateFormat}s.
	 * - Lines filled with {@code FORMAT_EXTRA} are added as extra data to output.
	 *
	 * @implSpec Every name we do not know is implicitly ignored.
	 */
	@Getter(AccessLevel.PUBLIC)
	@NotEmpty
	private List<String> format;

	@Getter(AccessLevel.PUBLIC)
	@NotEmpty
	private String[][] values;

	@Getter(AccessLevel.PUBLIC)
	private boolean onlySingles = false;

	/**
	 * Maps from Entity to the computed time-frame.
	 */
	@Getter(AccessLevel.PRIVATE)
	@JsonView(View.InternalCommunication.class)
	private Map<String, CDateSet> valuesResolved;

	@Getter(AccessLevel.PRIVATE)
	@JsonView(View.InternalCommunication.class)
	private String[] headers;

	/**
	 * Contains the uploaded additional data for each column for each entity.
	 * <p>
	 * Column -> Entity -> Value(s)
	 *
	 * @implNote FK: I would prefer to implement this as a guava table, but they cannot be deserialized with Jackson so we implement the Table manually.
	 */
	@Getter(AccessLevel.PRIVATE)
	@JsonView(View.InternalCommunication.class)
	private Map<String, Map<String, List<String>>> extra;

	public CQExternal(List<String> format, @NotEmpty String[][] values, boolean onlySingles) {
		this.format = format;
		this.values = values;
		this.onlySingles = onlySingles;
	}

	public boolean containsDates() {
		return format.stream().anyMatch(DateFormat.NAMES::contains);
	}

	@Override
	public QPNode createQueryPlan(QueryPlanContext context, ConceptQueryPlan plan) {
		if (valuesResolved == null) {
			throw new IllegalStateException("CQExternal needs to be resolved before creating a plan");
		}

		final String[] extraHeaders = Streams.zip(
													 Arrays.stream(headers),
													 format.stream(),
													 (header, format) -> format.equals(FORMAT_EXTRA) ? header : null
											 )
											 .filter(Objects::nonNull)
											 .toArray(String[]::new);

		if (onlySingles) {
			return createExternalNodeOnlySingle(context, plan, extraHeaders);
		}
		return createExternalNodeForList(context, plan, extraHeaders);

	}

	@Override
	public void collectRequiredQueries(Set<ManagedExecutionId> requiredQueries) {

	}

	private ExternalNode<String> createExternalNodeOnlySingle(QueryPlanContext context, ConceptQueryPlan plan, String[] extraHeaders) {
		// Remove zero element Lists and substitute one element Lists by containing String
		final Map<String, Map<String, String>> extraFlat = extra.entrySet().stream()
																 .collect(Collectors.toMap(
																		 Map.Entry::getKey,
																		 entityToRowMap -> entityToRowMap.getValue().entrySet().stream()
																										 .filter(headerToValue -> !headerToValue.getValue()
																																				.isEmpty())
																										 .collect(Collectors.toMap(
																												 Map.Entry::getKey,
																												 headerToValue -> headerToValue.getValue()
																																			   .get(0)
																										 ))
																 ));

		final Map<String, ConstantValueAggregator<String>> extraAggregators = new HashMap<>(extraHeaders.length);
		for (String extraHeader : extraHeaders) {
			// Just allocating, the result type is irrelevant here
			final ConstantValueAggregator<String> aggregator = new ConstantValueAggregator<>(null, null);
			extraAggregators.put(extraHeader, aggregator);
			plan.registerAggregator(aggregator);

		}

		return new ExternalNode<>(context.getStorage().getDataset().getAllIdsTable(), valuesResolved, extraFlat, extraHeaders, extraAggregators);
	}

	private ExternalNode<List<String>> createExternalNodeForList(QueryPlanContext context, ConceptQueryPlan plan, String[] extraHeaders) {
		final Map<String, ConstantValueAggregator<List<String>>> extraAggregators = new HashMap<>(extraHeaders.length);
		for (String extraHeader : extraHeaders) {
			// Just allocating, the result type is irrelevant here
			final ConstantValueAggregator<List<String>> aggregator = new ConstantValueAggregator<>(null, null);
			extraAggregators.put(extraHeader, aggregator);
			plan.registerAggregator(aggregator);
		}

		return new ExternalNode<>(
				context.getStorage().getDataset().getAllIdsTable(),
				valuesResolved,
				extra,
				extraHeaders,
				extraAggregators
		);
	}

	/**
	 * For each row try and collect all dates.
	 *
	 * @return Row -> Dates
	 */
	private static CDateSet[] readDates(String[][] values, List<String> format, DateReader dateReader) {
		final CDateSet[] out = new CDateSet[values.length];

		final List<DateFormat> dateFormats = format.stream()
												   .map(CQExternal::resolveDateFormat)
												   // Don't use Stream#toList to preserve null-values
												   .collect(Collectors.toList());


		/*
		 If no format is provided, put empty dates into output.
		 This indicates that no date context was provided and
		 the entries are not restricted by any date restriction,
		 but can also don't contribute to any date aggregation.
		 */
		if (dateFormats.stream().allMatch(Objects::isNull)) {
			// Initialize empty
			for (int row = 0; row < values.length; row++) {
				out[row] = CDateSet.createEmpty();
			}
			return out;
		}

		for (int row = 1; row < values.length; row++) {
			try {
				final CDateSet dates = CDateSet.createEmpty();

				// Collect all specified dates into a single set.
				for (int col = 0; col < dateFormats.size(); col++) {
					final DateFormat dateFormat = dateFormats.get(col);

					if (dateFormat == null) {
						continue;
					}
					dateFormat.readDates(values[row][col], dateReader, dates);
				}

				if (dates.isEmpty()) {
					continue;
				}

				if (out[row] == null) {
					out[row] = CDateSet.createEmpty();
				}

				out[row].addAll(dates);
			}
			catch (Exception e) {
				log.warn("Failed to parse Date from {}", row, e);
			}
		}

		return out;
	}

	@Override
	public void resolve(QueryResolveContext context) {
		headers = values[0];

		final ResolveStatistic resolved =
				resolveEntities(values, format,
								context.getNamespace().getStorage().getIdMapping(),
								context.getConfig().getIdColumns(),
								context.getConfig().getLocale().getDateReader(),
								onlySingles
				);

		if (resolved.getResolved().isEmpty()) {
			throw new ConqueryError.ExternalResolveEmptyError();
		}

		if (!resolved.getUnreadableDate().isEmpty()) {
			log.warn(
					"Could not read {} dates. Not resolved: {}",
					resolved.getUnreadableDate().size(),
					resolved.getUnreadableDate().subList(0, Math.min(resolved.getUnreadableDate().size(), 10))
			);
		}

		if (!resolved.getUnresolvedId().isEmpty()) {
			log.warn(
					"Could not resolve {} ids. Not resolved: {}",
					resolved.getUnresolvedId().size(),
					resolved.getUnresolvedId().subList(0, Math.min(resolved.getUnresolvedId().size(), 10))
			);
		}

		valuesResolved = resolved.getResolved();
		extra = resolved.getExtra();
	}

	@Data
	public static class ResolveStatistic {

		@JsonIgnore
		private final Map<String, CDateSet> resolved;

		/**
		 * Entity -> Column -> Values
		 */
		@JsonIgnore
		private final Map<String, Map<String, List<String>>> extra;

		private final List<String[]> unreadableDate;
		private final List<String[]> unresolvedId;

	}

	/**
	 * Helper method to try and resolve entities in values using the specified format.
	 */
	public static ResolveStatistic resolveEntities(@NotEmpty String[][] values, @NotEmpty List<String> format, EntityIdMap mapping, IdColumnConfig idColumnConfig, @NotNull DateReader dateReader, boolean onlySingles) {
		final Map<String, CDateSet> resolved = new HashMap<>();

		final List<String[]> unresolvedDate = new ArrayList<>();
		final List<String[]> unresolvedId = new ArrayList<>();

		// extract dates from rows
		final CDateSet[] rowDates = readDates(values, format, dateReader);

		// Extract extra data from rows by Row, to be collected into by entities
		// Row -> Column -> Value
		final Map<String, String>[] extraDataByRow = readExtras(values, format);

		final List<Function<String[], EntityIdMap.ExternalId>> readers = IdColumnUtil.getIdReaders(format, idColumnConfig.getIdMappers());

		// We will not be able to resolve anything...
		if (readers.isEmpty()) {
			return new ResolveStatistic(Collections.emptyMap(), Collections.emptyMap(), Collections.emptyList(), List.of(values));
		}

		// Entity -> Column -> Values
		final Map<String, Map<String, List<String>>> extraDataByEntity = new HashMap<>();

		// ignore the first row, because this is the header
		for (int rowNum = 1; rowNum < values.length; rowNum++) {

			final String[] row = values[rowNum];

			if (rowDates[rowNum] == null) {
				unresolvedDate.add(row);
				continue;
			}

			String resolvedId = tryResolveId(row, readers, mapping);

			if (resolvedId == null) {
				unresolvedId.add(row);
				continue;
			}

			//read the dates from the row
			resolved.put(resolvedId, rowDates[rowNum]);

			// Entity was resolved for row so we collect the data.
			if (extraDataByRow[rowNum] != null) {

				for (Map.Entry<String, String> entry : extraDataByRow[rowNum].entrySet()) {
					extraDataByEntity.computeIfAbsent(resolvedId, (ignored) -> new HashMap<>())
									 .computeIfAbsent(entry.getKey(), (ignored) -> new ArrayList<>())
									 .add(entry.getValue());
				}
			}
		}

		if (onlySingles) {
			// Check that there is at most one value per entity and per column
			final boolean alright = extraDataByEntity.values().stream()
													 .map(Map::values)
													 .flatMap(Collection::stream)
													 .allMatch(l -> l.size() <= 1);
			if (!alright) {
				throw new ConqueryError.ExternalResolveOnePerRowError();
			}
		}

		return new ResolveStatistic(resolved, extraDataByEntity, unresolvedDate, unresolvedId);
	}

	/**
	 * Try to extract a {@link com.bakdata.conquery.models.identifiable.mapping.EntityIdMap.ExternalId} from the row,
	 * then try to map it to an internal {@link com.bakdata.conquery.models.query.entity.Entity}
	 */
	private static String tryResolveId(String[] row, List<Function<String[], EntityIdMap.ExternalId>> readers, EntityIdMap mapping) {
		String resolvedId = null;

		for (Function<String[], EntityIdMap.ExternalId> reader : readers) {
			final EntityIdMap.ExternalId externalId = reader.apply(row);

			if (externalId == null) {
				continue;
			}

			String innerResolved = mapping.resolve(externalId);

			if (innerResolved == null) {
				continue;
			}

			// Only if all resolvable ids agree on the same entity, do we return the id.
			if (resolvedId != null && !innerResolved.equals(resolvedId)) {
				log.error("`{}` maps to different Entities", (Object) row);
				continue;
			}

			resolvedId = innerResolved;
		}
		return resolvedId;
	}

	/**
	 * Try and extract Extra data from input to be returned as extra-data in output.
	 * <p>
	 * Line -> ( Column -> Value )
	 */
	private static Map<String, String>[] readExtras(String[][] values, List<String> format) {
		final String[] names = values[0];
		final Map<String, String>[] extrasByRow = new Map[values.length];


		for (int line = 1; line < values.length; line++) {
			for (int col = 0; col < format.size(); col++) {
				if (!format.get(col).equals(FORMAT_EXTRA)) {
					continue;
				}


				if (extrasByRow[line] == null) {
					extrasByRow[line] = new HashMap<>(names.length);
				}

				extrasByRow[line].put(names[col], values[line][col]);
			}
		}


		return extrasByRow;
	}


	@Override
	public RequiredEntities collectRequiredEntities(QueryExecutionContext context) {
		return new RequiredEntities(valuesResolved.keySet());
	}

	@Override
	public List<ResultInfo> getResultInfos() {
		if (extra == null) {
			return Collections.emptyList();
		}
		List<ResultInfo> resultInfos = new ArrayList<>();
		for (int col = 0; col < format.size(); col++) {
			if (!format.get(col).equals(FORMAT_EXTRA)) {
				continue;
			}

			String column = headers[col];

			resultInfos.add(new SimpleResultInfo(column, onlySingles ?
														 ResultType.StringT.INSTANCE :
														 new ResultType.ListT(ResultType.StringT.INSTANCE)));
		}

		return resultInfos;
	}


	@JsonIgnore
	@ValidationMethod(message = "Values and Format are not of same width.")
	public boolean isAllSameLength() {
		final int expected = format.size();
		return Arrays.stream(values).mapToInt(a -> a.length).allMatch(v -> expected == v);
	}

	@JsonIgnore
	@ValidationMethod(message = "Headers are not unique")
	public boolean isHeadersUnique() {
		Set<String> uniqueNames = new HashSet<>();
		Set<String> duplicates = new HashSet<>();

		for (String header : values[0]) {
			if (!uniqueNames.add(header)) {
				duplicates.add(header);
			}
		}

		if (duplicates.isEmpty()) {
			return true;
		}

		log.error("Duplicate Headers {}", duplicates);

		return false;
	}

	/**
	 * Try to resolve a date format, return nothing if not possible.
	 */
	private static DateFormat resolveDateFormat(String name) {
		try {
			return DateFormat.valueOf(name);
		}
		catch (IllegalArgumentException e) {
			return null; // Does not exist
		}
	}
}
