package com.bakdata.conquery.models.auth.basic;

import java.security.SecureRandom;
import java.util.Date;
import java.util.Random;

import javax.annotation.Nullable;

import com.auth0.jwt.JWT;
import com.auth0.jwt.algorithms.Algorithm;
import com.auth0.jwt.exceptions.JWTDecodeException;
import com.bakdata.conquery.models.auth.web.DefaultAuthFilter;
import io.dropwizard.auth.oauth.OAuthCredentialAuthFilter;
import io.dropwizard.util.Duration;
import jakarta.inject.Named;
import jakarta.ws.rs.container.ContainerRequestContext;
import jakarta.ws.rs.core.HttpHeaders;
import lombok.experimental.UtilityClass;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.RandomStringUtils;
import org.apache.commons.lang3.time.DateUtils;
import org.apache.shiro.authc.AuthenticationToken;
import org.apache.shiro.authc.BearerToken;
import org.jvnet.hk2.annotations.Service;

@UtilityClass
@Slf4j
public class JWTokenHandler {

	private static final String PREFIX = "Bearer";
	private static final String OAUTH_ACCESS_TOKEN_PARAM = "access_token";

	private static final Random RANDOM_GEN = new SecureRandom();
	private static final int TOKEN_SECRET_LENGTH = 32;

	/**
	 * Creates a signed JWT token for the authentication with the
	 * {@link LocalAuthenticationRealm}.
	 */
	public String createToken(String username, Duration duration, String issuer, Algorithm algorithm) {
		Date issueDate = new Date();
		Date expDate = DateUtils.addMinutes(issueDate, Long.valueOf(duration.toMinutes()).intValue());
		String token = JWT.create().withIssuer(issuer).withSubject(username).withIssuedAt(issueDate).withExpiresAt(expDate).sign(algorithm);
		return token;
	}

	/**
	 * Code obtained from the Dropwizard project {@link OAuthCredentialAuthFilter}.
	 * <p>
	 * Parses a value of the `Authorization` header in the form of `Bearer
	 * a892bf3e284da9bb40648ab10`.
	 *
	 * @param header the value of the `Authorization` header
	 * @return a token
	 */
	@Nullable
	public static String extractTokenFromHeader(ContainerRequestContext request) {

		final String header = request.getHeaders().getFirst(HttpHeaders.AUTHORIZATION);

		if (header == null) {
			return null;
		}

		final int space = header.indexOf(' ');
		if (space <= 0) {
			return null;
		}

		final String method = header.substring(0, space);
		if (!PREFIX.equalsIgnoreCase(method)) {
			return null;
		}

		return header.substring(space + 1);
	}

	@Nullable
	private static String extractTokenFromQuery(ContainerRequestContext request) {
		// If Authorization header is not used, check query parameter where token can be
		// passed as well
		String credentials = request.getUriInfo().getQueryParameters().getFirst(OAUTH_ACCESS_TOKEN_PARAM);
		if (credentials != null) {
			return credentials;
		}
		return null;
	}

	/**
	 * Generate a random secret.
	 */
	public static String generateTokenSecret() {
		return RandomStringUtils.random(TOKEN_SECRET_LENGTH, 0, 0, false, false, null, RANDOM_GEN);
	}

	/**
	 * Tries to extract a JWT form a request according to <a href=
	 * "https://tools.ietf.org/html/rfc6750">https://tools.ietf.org/html/rfc6750</a>.
	 *
	 * @param request
	 * @return
	 */
	public static class JWTokenExtractor implements DefaultAuthFilter.TokenExtractor {

		@Override
		public AuthenticationToken apply(ContainerRequestContext request) {
			String token = null;
			String tokenHeader = extractTokenFromHeader(request);
			String tokenQuery = extractTokenFromQuery(request);
			if (tokenHeader == null && tokenQuery == null) {
				// No token could be parsed
				return null;
			}
			else if (tokenHeader != null && tokenQuery != null) {
				log.warn("There were tokens in the request header and query string provided, which is forbidden. See: https://tools.ietf.org/html/rfc6750#section-2");
				return null;
			}
			else if (tokenHeader != null) {
				log.trace("Extracted the request header token");
				token = tokenHeader;
			}
			else {
				log.trace("Extracted the query string token");
				token = tokenQuery;
			}

			try {
				JWT.decode(token);
				return new BearerToken(token);

			}
			catch (JWTDecodeException e) {
				return null;
			}
		}
	}
}
