package com.bakdata.conquery.models.datasets.concepts;

import javax.annotation.CheckForNull;
import javax.annotation.Nullable;

import com.bakdata.conquery.io.jackson.serializer.NsIdRef;
import com.bakdata.conquery.models.common.daterange.CDateRange;
import com.bakdata.conquery.models.datasets.Column;
import com.bakdata.conquery.models.datasets.Dataset;
import com.bakdata.conquery.models.events.Bucket;
import com.bakdata.conquery.models.identifiable.Labeled;
import com.bakdata.conquery.models.identifiable.ids.NamespacedIdentifiable;
import com.bakdata.conquery.models.identifiable.ids.specific.ValidityDateId;
import com.fasterxml.jackson.annotation.JsonBackReference;
import com.fasterxml.jackson.annotation.JsonIgnore;
import io.dropwizard.validation.ValidationMethod;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import lombok.ToString;
import lombok.extern.slf4j.Slf4j;

@Getter
@Setter
@NoArgsConstructor
@Slf4j
public class ValidityDate extends Labeled<ValidityDateId> implements NamespacedIdentifiable<ValidityDateId>, DaterangeSelect {

	@NsIdRef
	@Nullable
	private Column column;
	@NsIdRef
	@Nullable
	private Column startColumn;
	@NsIdRef
	@Nullable
	private Column endColumn;
	@JsonBackReference
	@ToString.Exclude
	@EqualsAndHashCode.Exclude
	private Connector connector;

	public static ValidityDate create(Column column) {
		final ValidityDate validityDate = new ValidityDate();
		validityDate.setColumn(column);
		return validityDate;
	}

	public static ValidityDate create(Column startColumn, Column endColumn) {
		final ValidityDate validityDate = new ValidityDate();
		validityDate.setColumn(startColumn);
		validityDate.setColumn(endColumn);
		return validityDate;
	}

	@Override
	public ValidityDateId createId() {
		return new ValidityDateId(connector.getId(), getName());
	}

	@CheckForNull
	public CDateRange getValidityDate(int event, Bucket bucket) {
		// I spent a lot of time trying to create two classes implementing single/multi-column valditiy dates separately.
		// JsonCreator was not happy, and I could not figure out why. This is probably the most performant implementation that's not two classes.

		if (getColumn() != null) {
			if (bucket.has(event, getColumn())) {
				return bucket.getAsDateRange(event, getColumn());
			}

			return null;
		}

		final Column startColumn = getStartColumn();
		final Column endColumn = getEndColumn();

		final boolean hasStart = bucket.has(event, startColumn);
		final boolean hasEnd = bucket.has(event, endColumn);

		if (!hasStart && !hasEnd) {
			return null;
		}

		final int start = hasStart ? bucket.getDate(event, startColumn) : Integer.MIN_VALUE;
		final int end = hasEnd ? bucket.getDate(event, endColumn) : Integer.MAX_VALUE;

		return CDateRange.of(start, end);
	}

	public boolean containsColumn(Column column) {
		return column.equals(getColumn()) || column.equals(getStartColumn()) || column.equals(getEndColumn());
	}

	@JsonIgnore
	@ValidationMethod(message = "ValidityDate is not for Connectors' Table.")
	public boolean isForConnectorsTable() {

		final boolean anyColumnNotForConnector =
				(startColumn != null && !startColumn.getTable().equals(connector.getTable()))
				|| (endColumn != null && !endColumn.getTable().equals(connector.getTable()));

		final boolean columnNotForConnector = column != null && !column.getTable().equals(connector.getTable());

		return !anyColumnNotForConnector && !columnNotForConnector;
	}

	@JsonIgnore
	@Override
	public Dataset getDataset() {
		return connector.getDataset();
	}

}
