import styled from "@emotion/styled";
import { faCopy } from "@fortawesome/free-regular-svg-icons";
import {
  ReactNode,
  createContext,
  memo,
  useCallback,
  useContext,
  useState,
} from "react";
import { useHotkeys } from "react-hotkeys-hook";
import { useSelector } from "react-redux";

import IconButton from "../button/IconButton";
import Modal from "../modal/Modal";

import { StateT } from "./reducers";

const initialState = {
  isOpen: false,
  setOpen: () => {},
};

const Context = createContext<ReturnType<typeof useContextValue>>(initialState);

const useContextValue = () => {
  const [isOpen, setOpen] = useState(false);

  return { isOpen, setOpen };
};

export const AboutProvider = ({ children }: { children: ReactNode }) => (
  <Context.Provider value={useContextValue()}>{children}</Context.Provider>
);

export const useAbout = () => {
  return useContext(Context);
};

const Grid = styled("div")`
  display: grid;
  grid-template-columns: auto 1fr;
  margin-bottom: 20px;
  gap: 5px 20px;
`;

const Version = styled("code")`
  font-size: 16px;
  font-weight: bold;
`;

const useVersion = () => {
  const backendGitDescribe = useSelector<StateT, string>(
    (state) => state.startup.config.version,
  );

  // TODO: GET THIS TO WORK WHEN BUILDING INSIDE A DOCKER CONTAINER
  // const frontendGitCommit = preval`
  //   const { execSync } = require('child_process');
  //   module.exports = execSync('git rev-parse --short HEAD').toString();
  // `;
  // const frontendGitTag = preval`
  //   const { execSync } = require('child_process');
  //   module.exports = execSync('git describe --all --exact-match \`git rev-parse HEAD\`').toString();
  // `;

  // THIS IS GETTING STATICALLY REPLACED USING "VITE DEFINE"
  const frontendTimestamp = `__BUILD_TIMESTAMP__`.replace(/"/g, "");
  const frontendGitDescribe = `__BUILD_GIT_DESCRIBE__`.replace(/"/g, "");

  return {
    backendGitDescribe,
    frontendTimestamp,
    frontendGitDescribe,
  };
};

export const About = memo(() => {
  const { isOpen, setOpen } = useAbout();
  const toggleOpen = useCallback(() => setOpen((open) => !open), [setOpen]);
  const { backendGitDescribe, frontendTimestamp, frontendGitDescribe } =
    useVersion();

  const copyVersionToClipboard = () => {
    navigator.clipboard.writeText(
      `BE: ${backendGitDescribe} FE: ${frontendGitDescribe}`,
    );
    setOpen(false);
  };

  useHotkeys("shift+?", toggleOpen, [toggleOpen]);

  if (!isOpen) return null;

  return (
    <Modal headline="Version" onClose={() => setOpen(false)}>
      <Grid>
        <div>Backend</div>
        <Version>{backendGitDescribe}</Version>
        <div>Frontend</div>
        <Version>
          {frontendGitDescribe} – {frontendTimestamp}
        </Version>
      </Grid>
      <IconButton frame icon={faCopy} onClick={copyVersionToClipboard}>
        Copy version info
      </IconButton>
    </Modal>
  );
});
